/*
 * Copyright 2008,2009,2010,2011,2012 Red Hat, Inc.
 *
 * This Program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This Program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this Program; if not, write to the
 *
 *   Free Software Foundation, Inc.
 *   59 Temple Place, Suite 330
 *   Boston, MA 02111-1307 USA
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>

#ifdef HAVE_DIRSRV_SLAPI_PLUGIN_H
#include <nspr.h>
#include <nss.h>
#include <dirsrv/slapi-plugin.h>
#else
#include <slapi-plugin.h>
#endif

#include <rpc/xdr.h>
#include <sss_nss_idmap.h>

#include "backend.h"
#include "back-shr.h"
#include "plugin.h"
#include "map.h"
#include "back-sch.h"

struct backend_search_filter_config {
	bool_t search_user;
	bool_t search_group;
	bool_t search_uid;
	bool_t search_gid;
	bool_t name_set;
	char   *name;
};

/* Check simple filter to see if it has
 * (cn|uid|uidNumber|gidNumber=<value>) or (objectClass=posixGroup)
 * Called by slapi_filter_apply(). */
static int
backend_search_filter_has_cn_uid(Slapi_Filter *filter, void *arg)
{
	struct backend_search_filter_config *config = arg;
	struct berval *bval;
	char *filter_type;
	int f_choice, rc;

	f_choice = slapi_filter_get_choice(filter);
	rc = slapi_filter_get_ava(filter, &filter_type, &bval);
	if ((LDAP_FILTER_EQUALITY == f_choice) && (0 == rc)) {
		if (0 == strcasecmp(filter_type, "uidNumber")) {
			config->search_uid = TRUE;
			config->name_set = TRUE;
		} else if (0 == strcasecmp(filter_type, "gidNumber")) {
			config->search_gid = TRUE;
			config->name_set = TRUE;
		} else if (0 == strcasecmp(filter_type, "uid")) {
			config->search_user = TRUE;
			config->name_set = TRUE;
		} else if (0 == strcasecmp(filter_type, "cn")) {
			config->name_set = TRUE;
		} else if (0 == strcasecmp(filter_type, "memberUid")) {
			config->name_set = TRUE;
		} else if ((0 == strcasecmp(filter_type, "objectClass")) &&
			   (0 == strcasecmp(bval->bv_val, "posixGroup"))) {
			config->search_group = TRUE;
		}

		if ((NULL == config->name) && config->name_set) {
			config->name = bval->bv_val;
		}
	}

	if ((config->search_uid ||
	     config->search_gid ||
	     config->search_user ||
	     config->search_group) && (config->name != NULL)) {
		return SLAPI_FILTER_SCAN_STOP;
	}
	return SLAPI_FILTER_SCAN_CONTINUE;
}

static Slapi_Entry *
backend_retrieve_user_entry_from_sssd(char *user_name, bool_t is_uid,
				 struct backend_set_data *set_data,
				 struct backend_search_cbdata *cbdata)
{
	struct passwd pwd, *result;
	Slapi_Entry *entry;
	int rc;
	enum sss_id_type id_type;
	char *sid_str, *dn;

	if (set_data->sssd_buffer == NULL) {
		return NULL;
	}

	if (is_uid) {
		rc = getpwuid_r(atoi(user_name), &pwd,
				set_data->sssd_buffer,
				set_data->sssd_buffer_len, &result);
	} else {
		rc = getpwnam_r(user_name, &pwd,
				set_data->sssd_buffer,
				set_data->sssd_buffer_len, &result);
	}
	if ((result == NULL) || (rc != 0)) {
		return NULL;
	}

	if (pwd.pw_uid < cbdata->sssd_min_id) {
		return NULL;
	}

	entry = slapi_entry_alloc();
	if (entry == NULL) {
		return NULL;
	}

	dn =  slapi_escape_filter_value(slapi_ch_smprintf("uid=%s,%s", pwd.pw_name,
						slapi_sdn_get_dn(set_data->container_sdn)), -1);
	if (dn == NULL) {
		slapi_entry_free(entry);
		return NULL;
	}

	slapi_entry_add_string(entry,
			       "objectClass", "top");
	slapi_entry_add_string(entry,
			       "objectClass", "posixAccount");
	slapi_entry_add_string(entry,
			       "objectClass", "extensibleObject");
	slapi_entry_add_string(entry,
			       "uid", pwd.pw_name);
	slapi_entry_attr_set_int(entry,
				 "uidNumber", pwd.pw_uid);
	slapi_entry_attr_set_int(entry,
				 "gidNumber", pwd.pw_gid);
	slapi_entry_add_string(entry,
			       "gecos", pwd.pw_gecos);
	if (strlen(pwd.pw_gecos) > 0) {
		slapi_entry_add_string(entry,
				       "cn", pwd.pw_gecos);
	} else {
		slapi_entry_add_string(entry,
				       "cn", pwd.pw_name);
	}

	slapi_entry_add_string(entry,
			       "homeDirectory", pwd.pw_dir);
	slapi_entry_add_string(entry,
			       "loginShell", pwd.pw_shell);

	slapi_entry_set_dn(entry, dn);

	rc = sss_nss_getsidbyid(pwd.pw_uid, &sid_str, &id_type);
	if ((rc == 0) && (sid_str != NULL)) {
		slapi_entry_add_string(entry, "objectClass", "ipaNTUserAttrs");
		slapi_entry_add_string(entry, "ipaNTSecurityIdentifier", sid_str);
		free(sid_str);
	}

	return entry;
}


static Slapi_Entry *
backend_retrieve_group_entry_from_sssd(char *group_name, bool_t is_gid,
				 struct backend_set_data *set_data,
				 struct backend_search_cbdata *cbdata)
{
	struct group grp, *result;
	Slapi_Entry *entry;
	int rc, i;
	enum sss_id_type id_type;
	char *sid_str, *dn;

	if (set_data->sssd_buffer == NULL) {
		return NULL;
	}

	if (is_gid) {
		rc = getgrgid_r(atoi(group_name), &grp,
				set_data->sssd_buffer,
				set_data->sssd_buffer_len, &result);
	} else {
		rc = getgrnam_r(group_name, &grp,
				set_data->sssd_buffer,
				set_data->sssd_buffer_len, &result);
	}
	if ((result == NULL) || (rc != 0)) {
		return NULL;
	}

	if (grp.gr_gid < cbdata->sssd_min_id) {
		return NULL;
	}

	entry = slapi_entry_alloc();
	if (entry == NULL) {
		return NULL;
	}

	dn =  slapi_escape_filter_value(slapi_ch_smprintf("cn=%s,%s", grp.gr_name,
					     slapi_sdn_get_dn(set_data->container_sdn)), -1);
	if (dn == NULL) {
		slapi_entry_free(entry);
		return NULL;
	}

	slapi_entry_add_string(entry,
			       "objectClass", "top");
	slapi_entry_add_string(entry,
			       "objectClass", "posixGroup");
	slapi_entry_add_string(entry,
			       "objectClass", "extensibleObject");
	slapi_entry_add_string(entry,
			       "cn", grp.gr_name);
	slapi_entry_attr_set_int(entry,
				 "gidNumber", grp.gr_gid);

	if (grp.gr_mem) {
		for (i=0; grp.gr_mem[i]; i++) {
			slapi_entry_add_string(entry, "memberUid", grp.gr_mem[i]);
		}
	}

	slapi_entry_set_dn(entry, dn);

	rc = sss_nss_getsidbyid(grp.gr_gid, &sid_str, &id_type);
	if ((rc == 0) && (sid_str != NULL)) {
		slapi_entry_add_string(entry, "objectClass", "ipaNTGroupAttrs");
		slapi_entry_add_string(entry, "ipaNTSecurityIdentifier", sid_str);
		free(sid_str);
	}

	return entry;
}

static void
backend_search_sssd_set_entry(struct backend_set_data *set_data,
			       struct backend_search_cbdata *cbdata,
			       Slapi_Entry *entry)
{
	char *ndn;
	if (entry) {
		slapi_entry_add_string(entry,
				       "schema-compat-origin", "sssd");
		cbdata->matched = TRUE;
		ndn = slapi_entry_get_ndn(entry);
		cbdata->n_entries++;
		if (cbdata->closest_match) {
			free(cbdata->closest_match);
		}
		cbdata->closest_match = strdup(ndn);

		/* Entry is created, cache it via map. 
		 * Next request will be served from the cache */

		/* This code is executed under read lock but we need a write lock */
		map_unlock();
		map_wrlock();
		backend_set_entry(cbdata->pb, entry, set_data);
		map_unlock();
		map_rdlock();
		slapi_entry_free(entry);
	}
}

/* Check filter for a component like (uid=<value>) and if found,
 * perform look up against SSSD and create entry based on that */
void
backend_search_sssd(struct backend_set_data *set_data,
		    struct backend_search_cbdata *cbdata)
{
	int result, rc;
	Slapi_Entry *entry;
	struct backend_search_filter_config config = {FALSE, FALSE, FALSE, FALSE, FALSE, NULL};

	/* There was no match but we asked to check SSSD */
	/* First, we search the filter to see if it includes cn|uid=<value> check */
	result = slapi_filter_apply(cbdata->filter,
		       backend_search_filter_has_cn_uid, &config, &rc);
	if ((result != SLAPI_FILTER_SCAN_STOP) ||
	    (NULL == config.name)) {
		return;
	}

	/* Drop irrelevant requests. Each set only works with a single type */
	if ((cbdata->check_sssd == SCH_SSSD_GROUP) &&
	    (config.search_uid || config.search_user)) {
		return;
	}

	if ((cbdata->check_sssd == SCH_SSSD_USER) &&
	    (config.search_gid || config.search_group)) {
		return;
	}

	if ((config.search_gid || config.search_uid) &&
	    (atol(config.name) < set_data->sssd_min_id)) {
		return;
	}

	if ((config.search_group || config.search_gid) &&
	    (NULL != config.name)) {
		entry = backend_retrieve_group_entry_from_sssd(config.name,
						config.search_gid, set_data, cbdata);
		backend_search_sssd_set_entry(set_data, cbdata, entry);
		return;
	}

	if ((config.search_user || config.search_uid) &&
	    (NULL != config.name)) {
		entry = backend_retrieve_user_entry_from_sssd(config.name,
						config.search_uid, set_data, cbdata);
		backend_search_sssd_set_entry(set_data, cbdata, entry);
		return;
	}
}
